/*********************************************************************
 * Software License Agreement (BSD License)
 *
 *  A button + audio mobile robot user interface
 *  Copyright (c) 2011, Bob Mottram
 *  fuzzgun@gmail.com
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *********************************************************************/

#include <ros/ros.h>
#include <stdio.h>
#include <stdlib.h>
#include <fstream>
#include "serialport.cpp"
#include <geometry_msgs/Twist.h>
#include <geometry_msgs/Pose.h>
#include <geometry_msgs/PoseArray.h>
#include <geometry_msgs/PoseWithCovarianceStamped.h>
#include <move_base_msgs/MoveBaseAction.h>
#include <actionlib/client/simple_action_client.h>
#include <tf/transform_broadcaster.h>
#include "buttonui.h"

using namespace std;

typedef actionlib::SimpleActionClient<move_base_msgs::MoveBaseAction> MoveBaseClient;

// Has the initial pose been set
bool initial_pose_set = false;

// sounds when buttons are pressed
std::string button_sound[] = {
  "Started", "Stopped", "%s remembered", "%s", "Location set to %s", "Going to %s"
};

// commandline instructions when buttons are pressed
std::string button_command[] = {
  "","","","",""
};

std::string enabled_sound = "User interface enabled";
std::string arrived_goal_sound = "Arrived at %s";
std::string failed_goal_sound = "I can't get to %s";

// Topic subscribed to in order to obtain poses, typically /acml_pose
std::string pose_topic = "";

// Topic to publish pose array
std::string pose_array_topic = "/particlecloud";

// Topic to publish the initial pose to
std::string initial_pose_topic = "/initialpose";

// File used to store poses
std::string placemarks_filename="places.txt";

// command to be used for speech
std::string speak_command = "espeak";

// The current pose estimate
geometry_msgs::PoseWithCovarianceStamped current_pose;
bool pose_received = false;

// place names
int current_place_name_index = -1;
std::vector<std::string> place_names;

// Is the robot moving to a goal
bool moving_to_goal = false;
MoveBaseClient *ac = NULL;

// A file containing the current pose information
// This allows the robot to recover its pose after being switched off
std::string current_pose_filename="";

// Initial pose translational uncertainty in metres
float initial_pose_radius = 0.1f;

// Initial pose angular uncertainty in radians
float initial_pose_orientation_variance = M_PI/12.0;

ros::Publisher motor_pub;
ros::Publisher initial_pose_pub;
ros::Publisher pose_array_pub;
geometry_msgs::Twist cmd;

bool started = false;
bool initialised = false;

// What time was the learn button previously pressed ?
ros::Time previous_learn_button_press;

// Which button was previously pressed ?
int previous_button = -1;


int say(const char * text)
{
  char str[255];
  sprintf((char*)str,"%s \"%s\" --stdout | aplay",speak_command.c_str(),text);
  ROS_INFO((char*)str);
  if (system(str)) {
    return 1;
  }
  return 0;
}

int say(std::string text)
{
  return say(text.c_str());
}

bool is_enabled()
{
  return true;
}

bool valid_pose()
{
  const double max = 99999;
  if ((current_pose.pose.pose.position.x>-max) &&
      (current_pose.pose.pose.position.x<max) &&
      (current_pose.pose.pose.position.y>-max) &&
      (current_pose.pose.pose.position.y<max) &&
      (current_pose.pose.pose.orientation.z>-max) &&
      (current_pose.pose.pose.orientation.z<max) &&
      (current_pose.pose.pose.orientation.w>-max) &&
      (current_pose.pose.pose.orientation.w<max)) {
    return true;
  }
  return false;
}

/*
 * \brief Saves the current pose to a file
 */
void save_current_pose(std::string filename)
{
  if (valid_pose()) {
    FILE * fp = fopen(filename.c_str(),"w");
    if (fp != NULL) {
      ROS_INFO("Saving current pose %.3f, %.3f, %.3f, %.3f",
	       current_pose.pose.pose.position.x,
	       current_pose.pose.pose.position.y,
	       current_pose.pose.pose.orientation.z,
	       current_pose.pose.pose.orientation.w);
      fprintf(fp,"%f,%f,%f,%f",
	      current_pose.pose.pose.position.x,
	      current_pose.pose.pose.position.y,
	      current_pose.pose.pose.orientation.z,
	      current_pose.pose.pose.orientation.w);
      fclose(fp);
    }
  }
}

int read_placemarks(
		    std::vector<std::string> &places,
		    std:: vector<float> &poses)
{
  int idx=0,index=-1,record_index=0;
  FILE * fp = fopen(placemarks_filename.c_str(), "r");
  if (fp != NULL) {
    std::string str="";
    while (!feof(fp)) {
      char c = fgetc(fp);
      if (c == ',') {
	if (idx == 0) {
	  places.push_back(str);
	  if (current_place_name_index > -1) {
	    if (str==place_names[current_place_name_index]) {
	      index = record_index;
	    }
	  }
	}
	else {
	  poses.push_back((float)atof(str.c_str()));
	}
	str="";
	idx++;
      }
      else {
	if (c == '\n') {
	  poses.push_back((float)atof(str.c_str()));
	  idx = 0;
	  record_index++;
	  str="";
	}
	else {
	  str += c;
	}
      }
    }
    if (str!="") {
      poses.push_back((float)atof(str.c_str()));
    }
    fclose(fp);                
  }
  return index;
}

void learn_pose()
{
  if (valid_pose()) {

    if (is_enabled()) {
      if (placemarks_filename=="") {
        ROS_WARN("No placemarks filename specified");
      }
      if (!pose_received) {
	ROS_WARN("Pose not received");
      }

      if ((pose_received) && (placemarks_filename != "") && (current_place_name_index > -1)) {
	bool saved = false;

	if (button_command[BUTTON_LEARN] != "") {
	  if (system(button_command[BUTTON_LEARN].c_str())!=0) {
	    ROS_INFO("Running command:\n%s",button_command[BUTTON_LEARN].c_str());
	  }
	}

	std::vector<std::string> places;
	std:: vector<float> poses;

	int index = read_placemarks(places, poses);

	// save to file
	if (index == -1) {
	  ROS_INFO("New Placemark");
	  // append or write new place to file
	  FILE * fp = fopen(placemarks_filename.c_str(), "a");
	  if (fp == NULL) fp = fopen(placemarks_filename.c_str(), "w");
	  if (fp != NULL) {
	    ROS_INFO("%s, %.3f, %.3f, %.3f, %.3f",
		     place_names[current_place_name_index].c_str(),
		     current_pose.pose.pose.position.x,
		     current_pose.pose.pose.position.y,
		     current_pose.pose.pose.orientation.z,
		     current_pose.pose.pose.orientation.w);
	    fprintf(fp,"%s, %f, %f, %f, %f\n",
		    place_names[current_place_name_index].c_str(),
		    current_pose.pose.pose.position.x,
		    current_pose.pose.pose.position.y,
		    current_pose.pose.pose.orientation.z,
		    current_pose.pose.pose.orientation.w);
	    fclose(fp);
	    saved = true;
	  }
	}
	else {
	  // alter an existing place
	  ROS_INFO("Existing Placemark");
	  poses[index*4] = current_pose.pose.pose.position.x;
	  poses[index*4+1] = current_pose.pose.pose.position.y;
	  poses[index*4+2] = current_pose.pose.pose.orientation.z;
	  poses[index*4+3] = current_pose.pose.pose.orientation.w;
	  FILE * fp = fopen(placemarks_filename.c_str(), "w");
	  if (fp != NULL) {
	    for (int i = 0; i < (int)places.size(); i++) {
	      fprintf(fp, "%s, %f, %f, %f, %f\n", places[i].c_str(), poses[i*4], poses[i*4+1], poses[i*4+2], poses[i*4+3]);
	    }
	    fclose(fp);
	    saved = true;
	  }
	}

	if ((saved) && (button_sound[BUTTON_LEARN] !="")) {
	  char saystr[255];
	  sprintf((char*)saystr,button_sound[BUTTON_LEARN].c_str(),place_names[current_place_name_index].c_str());
	  say(saystr);
 	  ROS_INFO("learn_pose: %s",saystr);
	}
      }
    }
  }
}


/*!
 * \brief Select the next place name
 */
void select_next_place_name()
{
  if (is_enabled()) {
    if (place_names.size() > 0) {
      // increment the place name
      current_place_name_index++;
      if (current_place_name_index >= (int)place_names.size()) {
	current_place_name_index = 0;
      }
      // say the place name
      say(place_names[current_place_name_index]);
    }
  }
}

/*!
 * \brief Set the current pose
 */
void set_current_pose()
{
  if (is_enabled()) {
    if (place_names.size() > 0) {
      std::vector<std::string> places;
      std::vector<float> poses;
      int index = read_placemarks(places, poses);
      if (index > -1) {
	// create an initial pose
	geometry_msgs::PoseWithCovarianceStamped p = current_pose;                
	p.header.frame_id = "/map";
	p.header.stamp = ros::Time::now();
	p.pose.pose.position.x = poses[index*4];
	p.pose.pose.position.y = poses[index*4+1];
	p.pose.pose.orientation.z = poses[index*4+2];
	p.pose.pose.orientation.w = poses[index*4+3];
	for (int i = 0; i < (int)p.pose.covariance.size(); i++) {
	  p.pose.covariance[i]=0;
	}
	p.pose.covariance[6*0+0] = initial_pose_radius * initial_pose_radius;
	p.pose.covariance[6*1+1] = initial_pose_radius * initial_pose_radius;
	p.pose.covariance[6*3+3] = initial_pose_orientation_variance * initial_pose_orientation_variance;

	if (initial_pose_topic !="") {
	  initial_pose_pub.publish(p);
	}

	char saystr[255];
	sprintf((char*)saystr,button_sound[BUTTON_SET_POSE].c_str(),place_names[current_place_name_index].c_str());
	say(saystr);
      }
      else {
	ROS_WARN("No placemark found");
      }
    }
  }
}


void go()
{
  if ((current_place_name_index > -1) && (!moving_to_goal)) {
    std::vector<std::string> places;
    std::vector<float> poses;

    int index = read_placemarks(places, poses);

    if (index > -1) {

      move_base_msgs::MoveBaseGoal goal;

      goal.target_pose.header.frame_id = "/map";
      goal.target_pose.header.stamp = ros::Time::now();

      goal.target_pose.pose.position.x = poses[index*4];
      goal.target_pose.pose.position.y = poses[index*4+1];
      goal.target_pose.pose.orientation.z = poses[index*4+2];
      goal.target_pose.pose.orientation.w = poses[index*4+3];

      ROS_INFO("Sending goal");

      //tell the action client that we want to spin a thread by default
      if (ac==NULL) ac = new MoveBaseClient("move_base", true);

      //wait for the action server to come up
      while(!ac->waitForServer(ros::Duration(5.0))){
	ROS_INFO("Waiting for the move_base action server to come up");
      }
      ac->sendGoal(goal);

      moving_to_goal = true;

      char saystr[255];
      sprintf((char*)saystr,button_sound[BUTTON_GO].c_str(),place_names[current_place_name_index].c_str());
      say(saystr);
    }
  }
}

/*!
 * \brief callback when a new pose estimate is received
 * \param ptr pose
 */
void poseCallback(
		  const geometry_msgs::PoseWithCovarianceStamped::ConstPtr& ptr)
{
  if (initialised) {
    geometry_msgs::PoseWithCovarianceStamped p = *ptr;
    current_pose = p;
    ROS_INFO("Pose received %.3f, %.3f, %.3f, %.3f", p.pose.pose.position.x,p.pose.pose.position.y,p.pose.pose.orientation.z,p.pose.pose.orientation.w);
    if (initial_pose_set) save_current_pose(current_pose_filename);
    pose_received = true;
  }
}

/*!
 * \brief sets all velocities to zero
 */
void stop_motors()
{
  if (initialised) {
    cmd.linear.x = 0;
    cmd.linear.y = 0;
    cmd.linear.z = 0;
    cmd.angular.x = 0;
    cmd.angular.y = 0;
    cmd.angular.z = 0;
  }

  ROS_INFO("Motors stopped");
}

/*
 * \brief Loads the initial pose from file.  This allows the robot to know its initial location after being switched on.
 */
void load_initial_pose(std::string filename)
{
  FILE * fp = fopen(filename.c_str(),"r");
  if (fp != NULL) {
    geometry_msgs::PoseWithCovarianceStamped p = current_pose;
    p.header.stamp = ros::Time::now();
    p.header.frame_id = "/map";
    int index=0;
    std::string str="";
    while (!feof(fp)) {
      char c = fgetc(fp);
      if ((c == ',') || (c == '\n')) {
	if (str != "") {
	  switch(index) {
	  case 0: { p.pose.pose.position.x = atof(str.c_str()); break; };
	  case 1: { p.pose.pose.position.y = atof(str.c_str()); break; };
	  case 2: { p.pose.pose.orientation.z = atof(str.c_str()); break; };
	  case 3: { p.pose.pose.orientation.w = atof(str.c_str()); break; };
	  }
	  index++;
	  str="";
	}
      }
      else {
	str += c;
      }
    }
    if (str != "") {
      p.pose.pose.orientation.w = atof(str.c_str());
    }

    p.pose.covariance[6*0+0] = initial_pose_radius * initial_pose_radius;
    p.pose.covariance[6*1+1] = initial_pose_radius * initial_pose_radius;
    p.pose.covariance[6*3+3] = initial_pose_orientation_variance * initial_pose_orientation_variance;

    fclose(fp);
    if (initial_pose_topic !="") {
      ROS_INFO("Publishing initial pose %s %.3f, %.3f,%.3f,%.3f to %s",
	       p.header.frame_id.c_str(),
	       p.pose.pose.position.x,
	       p.pose.pose.position.y,
	       p.pose.pose.orientation.z,
	       p.pose.pose.orientation.w,
	       initial_pose_topic.c_str());
      initial_pose_pub.publish(p);
    }
    current_pose = p;
  }
  else {
    ROS_WARN("No known current location");
  }
}

/*!
 * \brief loads place names from a comma separated string.  eg. Kitchen table,Lounge,Hallway
 * \param placenames string containing place names
 */
void load_place_names(std::string placenames)
{
  place_names.clear();
  if (placenames != "") {
    const char * place_names_str = placenames.c_str();
    std::string str="";
    for (int i = 0; i < (int)placenames.length(); i++) {
      char c = place_names_str[i];
      if (c != ',') {
	str += c;
      }
      else {
	if (str != "") {
	  place_names.push_back(str);
	  str = "";
	}
      }
    }
    if (str != "") {
      place_names.push_back(str);
    }
  }
}

// Receive button presses from the Arduino
int get_button_press(
		     SerialPort &serial,
		     unsigned timeout_mS)
{
  const int bytes = 4;
  int index = -1;
  unsigned char comBuffer[bytes];
  int returned_bytes = serial.In(comBuffer, bytes, timeout_mS);
  if (returned_bytes > 0) {
    if (((char)comBuffer[0]>='0') && ((char)comBuffer[0]<='9')) {
      index = atoi((char*)comBuffer);
      int index2 = 0;
      for (index2=0;index2<sizeof(button_index)/sizeof(int);index2++) {
	if (button_index[index2]==index) break;
      }
      ROS_INFO("Button %d (%s) pressed", index,button_name[index2].c_str());
      index = index2;
    }
  }
  return index;
}

// Do stuff when buttons are pressed
void update(bool * prev_enable, bool * prev_started, int index)
{
  switch(index) {
  case BUTTON_START: {
    if (started) {
      ROS_INFO("Go button pressed");
      go();
    }
    else {
      if (previous_button==button_index[BUTTON_SET_POSE]) {
	// Pressing start moves to the next location
	current_place_name_index++;
	if (current_place_name_index >= (int)place_names.size()) {
	  current_place_name_index = 0;
	}
      }
      started = true;

      // load the initial pose
      if (button_sound[BUTTON_START] !="") say(button_sound[BUTTON_START]);
      if (button_command[BUTTON_START] != "") {
	ROS_INFO("Running command:\n%s",button_command[BUTTON_START].c_str());
	system(button_command[BUTTON_START].c_str());
      }

      ROS_INFO("Start button pressed");
    }
    break;
  }
  case BUTTON_STOP: {
    if (started) {
      started = false;

      if (button_sound[BUTTON_STOP] !="") say(button_sound[BUTTON_STOP]);
      if (button_command[BUTTON_STOP] != "") {
	ROS_INFO("Running command:\n%s",button_command[BUTTON_STOP].c_str());
	system(button_command[BUTTON_STOP].c_str());
      }

      ROS_INFO("Stop button pressed");
    }
    break;
  }
  case BUTTON_LEARN: {
    if (started) {
      ROS_INFO("Learn button pressed");
      if ((ros::Time::now()-previous_learn_button_press).toSec()<=3) {
	learn_pose();
      }
      previous_learn_button_press = ros::Time::now();
    }
    break;
  }
  case BUTTON_PLACE: {
    if (started) {
      ROS_INFO("Selecting place name");
      select_next_place_name();
    }
    break;
  }
  case BUTTON_SET_POSE: {
    if (started) {
      ROS_INFO("Set current pose");
      set_current_pose();
    }
    break;
  }
  }


  // has buttonui been enabled ?
  bool en = is_enabled();

  if ((pose_received) && (!initial_pose_set)) {
    load_initial_pose(current_pose_filename);
    initial_pose_set = true;
  }

  if ((en) && (!*prev_enable)) {
    if (enabled_sound !="") say(enabled_sound);
  }

  if (moving_to_goal) {
    if (ac->getState() == actionlib::SimpleClientGoalState::SUCCEEDED) {
      char saystr[255];
      sprintf((char*)saystr,arrived_goal_sound.c_str(),place_names[current_place_name_index].c_str());
      ROS_INFO((char*)saystr);
      say(saystr);
      // Proceed to the next location
      current_place_name_index++;
      if (current_place_name_index>=place_names.size()) current_place_name_index=0;
      moving_to_goal = false;
    }
    if ((ac->getState() == actionlib::SimpleClientGoalState::LOST) ||
	(ac->getState() == actionlib::SimpleClientGoalState::REJECTED) ||
	(ac->getState() == actionlib::SimpleClientGoalState::ABORTED)) {
      char saystr[255];
      sprintf((char*)saystr,failed_goal_sound.c_str(),place_names[current_place_name_index].c_str());
      ROS_INFO((char*)saystr);
      say(saystr);
      moving_to_goal = false;
    }
  }

  // if joystick control has started and the enable flag goes low
  // then ensure that the motors are stopped
  if ((!en) && (*prev_enable) && (started)) {
    started = false;
    stop_motors();
  }
  if ((en) && (!started) && (*prev_started)) {
    stop_motors();
  }
  *prev_enable = en;
  *prev_started = started;
    
}

int main(int argc, char* argv[])
{
  ros::init(argc, argv, "buttonui");
  ros::NodeHandle n;
  ros::NodeHandle nh("~");

  sleep(8); // pause to wait for the sound client to connect

  bool enable = false;
  bool prev_enable = false;
  bool prev_started = false;
  nh.getParam("enable", enable);

  std::string name = "cmd_vel";
  nh.getParam("name", name);
  std::string topic_path = "buttonui/";
  nh.getParam("topic_path", topic_path);

  // sounds
  std::string sound_str = "";
  nh.getParam("startbuttonsound", sound_str);
  if (sound_str != "") button_sound[BUTTON_START] = sound_str;
  sound_str = "";
  nh.getParam("stopbuttonsound", sound_str);
  if (sound_str != "") button_sound[BUTTON_STOP] = sound_str;
  sound_str = "";
  nh.getParam("learnbuttonsound", sound_str);
  if (sound_str != "") button_sound[BUTTON_LEARN] = sound_str;
  sound_str = "";
  nh.getParam("gobuttonsound", sound_str);
  if (sound_str != "") button_sound[BUTTON_GO] = sound_str;
  sound_str = "";
  nh.getParam("enablesound", sound_str);
  if (sound_str != "") enabled_sound = sound_str;


  // start and stop button digital inputs on the interface kit
  int i=-1;
  std::string str="";
  nh.getParam("startbuttonindex", i);
  if (i>-1) button_index[BUTTON_START] = i;
  nh.getParam("startbuttoncommand", str);
  if (str!="") button_command[BUTTON_START] = str;
  i=-1;
  nh.getParam("stopbuttonindex", i);
  if (i>-1) button_index[BUTTON_STOP] = i;
  str="";
  nh.getParam("stopbuttoncommand", str);
  if (str!="") button_command[BUTTON_STOP] = str;

  // Button for learning new locations/poses
  std::string placenames = "";
  nh.getParam("pose_topic", pose_topic);
  nh.getParam("pose_array_topic", pose_array_topic);
  nh.getParam("initial_pose_topic", initial_pose_topic);
  nh.getParam("current_pose_filename", current_pose_filename);
  nh.getParam("placemarks", placemarks_filename);
  nh.getParam("placenames", placenames);
  i=-1;
  nh.getParam("setposebuttonindex", i);
  if (i>-1) button_index[BUTTON_SET_POSE] = i;
  i=-1;
  nh.getParam("placebuttonindex", i);
  if (i>-1) button_index[BUTTON_PLACE] = i;
  i=-1;
  nh.getParam("learnbuttonindex", i);
  if (i>-1) button_index[BUTTON_LEARN] = i;
  i=-1;
  nh.getParam("gobuttonindex", i);
  if (i>-1) button_index[BUTTON_START] = i;
  str="";
  nh.getParam("learnbuttoncommand", str);
  if (str!="") button_command[BUTTON_LEARN] = str;

  // pose uncertainty
  double val=0;
  nh.getParam("initial_pose_radius", val);
  if (val > 0.001) initial_pose_radius = val;
  nh.getParam("initial_pose_orientation_variance", val);
  if (val > 0.001) initial_pose_orientation_variance = val*M_PI/180.0;

  // speak command
  std::string speakcommand = "";
  nh.getParam("speakcommand", speakcommand);
  if (speakcommand != "") speak_command = speakcommand;

  // get the device to read from
  std::string device_name = "";
  nh.getParam("device", device_name);
  if (device_name == "") device_name = "/dev/ttyUSB0";

  // Communication with Arduino
  int baud_rate = 9600;
  std::string baud_rate_str = "";
  nh.getParam("baud", baud_rate_str);
  if (baud_rate_str!="") {
    baud_rate = atoi(baud_rate_str.c_str());
  }

  int data_bits = 8;
  std::string data_bits_str = "";
  nh.getParam("databits", data_bits_str);
  if (data_bits_str!="") {
    data_bits = atoi(data_bits_str.c_str());
  }

  int stop_bits = 1;
  std::string stop_bits_str = "";
  nh.getParam("stopbits", data_bits_str);
  if (stop_bits_str!="") {
    stop_bits = atoi(stop_bits_str.c_str());
  }


  // subscribe to poses
  ros::Subscriber poses_sub;
  if (pose_topic != "") {
    poses_sub = n.subscribe(pose_topic, 1, poseCallback);
  }

  // initial pose publisher
  if (initial_pose_topic != "") {
    initial_pose_pub = n.advertise<geometry_msgs::PoseWithCovarianceStamped>(initial_pose_topic, 10);
  }

  // pose array publisher
  if (pose_array_topic != "") {
    pose_array_pub = n.advertise<geometry_msgs::PoseArray>(pose_array_topic, 10);
  }

  // halt the motors
  stop_motors();

  if (!enable) {
    ROS_INFO("Waiting for enable parameter to be set");
  }
  else {
    if (enabled_sound !="") say(enabled_sound);
  }

  // load place names
  load_place_names(placenames);

  // load the initial pose
  load_initial_pose(current_pose_filename);


  SerialPort serial;
  if (serial.Open(device_name) == 0) {
    serial.Initialise(baud_rate, data_bits, stop_bits, PARITY_NONE);

    ROS_INFO("Connected to %s", device_name.c_str());

    ros::Rate loop_rate(10);
    initialised = true;
    while (n.ok()) {
      update(&prev_enable, &prev_started, get_button_press(serial,100));
      ros::spinOnce();
      loop_rate.sleep();
    }

    ROS_INFO("Closing %s", device_name.c_str());
  }
  else {
    ROS_WARN("Can't open %s", device_name.c_str());
  }

  if (ac!=NULL) delete ac;

  return 0;
}

